# 舞萌DX
# 标题服务器通讯实现

import zlib
import hashlib
import httpx
from loguru import logger
import random
import time
from ctypes import c_int32
from Crypto.Cipher import AES
from Crypto.Util.Padding import pad, unpad
from Config import *
import certifi

# 舞萌DX 2024
# omg it's leaking
AesKey = "n7bx6:@Fg_:2;5E89Phy7AyIcpxEQ:R@"
AesIV = ";;KjR1C3hgB1ovXa"
ObfuscateParam = "BEs2D5vW"

# 2025
AesKey = "a>32bVP7v<63BVLkY[xM>daZ1s9MBP<R"
AesIV = "d6xHIKq]1J]Dt^ue"
ObfuscateParam = "B44df8yT"

class SDGBApiError(Exception):
    pass
    
class SDGBRequestError(SDGBApiError):
    pass

class SDGBResponseError(SDGBApiError):
    pass

class AESPKCS7:
    # 实现了 maimai 通讯所用的 AES 加密的类
    def __init__(self, key: str, iv: str):
        self.key = key.encode('utf-8')
        self.iv = iv.encode('utf-8')
        self.mode = AES.MODE_CBC
    # 加密
    def encrypt(self, content) -> bytes:
        # if content is str, convert to bytes
        #if isinstance(content, str):
        #    encodedData = content.encode('utf-8')
        cipher = AES.new(self.key, self.mode, self.iv)
        content_padded = pad(content, AES.block_size)
        encrypted_bytes = cipher.encrypt(content_padded)
        return encrypted_bytes
    # 解密
    def decrypt(self, encrypted_content: bytes) -> str:
        cipher = AES.new(self.key, self.mode, self.iv)
        decrypted_padded = cipher.decrypt(encrypted_content)
        decrypted = unpad(decrypted_padded, AES.block_size)
        return decrypted

def getSDGBApiHash(api):
    # API 的 Hash 的生成
    # 有空做一下 Hash 的彩虹表？
    return hashlib.md5((api+"MaimaiChn"+ObfuscateParam).encode()).hexdigest()

def apiSDGB(data:str, targetApi:str, userAgentExtraData:str, noLog:bool=False, timeout:int=5):
    """
    舞萌DX 2024 API 通讯用函数
    :param data: 请求数据
    :param targetApi: 使用的 API
    :param userAgentExtraData: UA 附加信息，机台相关则为狗号（如A63E01E9564），用户相关则为 UID
    :param noLog: 是否不记录日志
    """
    maxRetries = 3
    agentExtra = str(userAgentExtraData)
    aes = AESPKCS7(AesKey, AesIV)
    # Begin Build
    requestDataFinal = aes.encrypt(zlib.compress(data.encode('utf-8')))
    # End Build
    endpoint = "https://maimai-gm.wahlap.com:42081/Maimai2Servlet/"
    if not noLog:
        logger.debug(f"开始请求 {targetApi}，以 {data}")

    retries = 0
    while retries < maxRetries:
        try:
            if useProxy:
                # 使用代理
                logger.debug("使用代理")
                httpClient = httpx.Client(proxy=proxyUrl, verify=False)
            else:
                # 不使用代理
                logger.debug("不使用代理")
                httpClient = httpx.Client(verify=False)
            responseDataRaw = httpClient.post(
                url=endpoint + getSDGBApiHash(targetApi),
                headers={
                    "User-Agent": f"{getSDGBApiHash(targetApi)}#{agentExtra}",
                    "Content-Type": "application/json",
                    "Mai-Encoding": "1.50",
                    "Accept-Encoding": "",
                    "Charset": "UTF-8",
                    "Content-Encoding": "deflate",
                    "Expect": "100-continue"
                },
                content=requestDataFinal,
                # 经测试，加 Verify 之后速度慢好多，因此建议选择性开
                #verify=certifi.where(),
                #verify=False,
                timeout=timeout
            )

            if not noLog:
                logger.info(f"{targetApi} 请求结果: {responseDataRaw.status_code}")

            if responseDataRaw.status_code == 200:
                logger.debug("200 OK!")
            else:
                errorMessage = f"请求失败: {responseDataRaw.status_code}"
                logger.error(errorMessage)
                raise SDGBRequestError(errorMessage)

            responseContentRaw = responseDataRaw.content


            try:
                responseContentDecrypted = aes.decrypt(responseContentRaw)
                logger.debug(f"成功解密响应！")
            except:
                logger.warning(f"解密失败，得到的原始响应: {responseContentRaw}")
                raise SDGBResponseError("解密失败")
            try:
                responseDataFinal = zlib.decompress(responseContentDecrypted)
                logger.debug("成功解压响应！")
            except:
                logger.warning(f"无法解压，解密的原始响应: {responseContentDecrypted}")
                raise SDGBResponseError("解压失败")

            
            if not noLog:
                logger.debug(f"响应: {responseContentDecrypted}")
            return responseContentDecrypted
        
        # 异常处理
        except SDGBRequestError as e:
            # 请求格式错误，不需要重试
            raise SDGBRequestError("请求格式错误")
        except SDGBResponseError as e:
            # 响应解析错误，这种有一定可能是我们的问题，所以只重试一次
            logger.warning(f"将重试一次 Resp Err: {e}")
            retries += 2
            time.sleep(2)
        except Exception as e:
            # 其他错误，重试多次
            logger.warning(f"将开始重试请求. {e}")
            retries += 1
            time.sleep(2)

    raise SDGBApiError("重试多次仍然无法成功请求服务器")

def calcPlaySpecial():
    """使用 c_int32 实现的 SpecialNumber 算法"""
    rng = random.SystemRandom()
    num2 = rng.randint(1, 1037933) * 2069
    num2 += 1024  #GameManager.CalcSpecialNum()
    num2 = c_int32(num2).value
    result = c_int32(0)
    for _ in range(32):
        result.value <<= 1
        result.value += num2 & 1
        num2 >>= 1
    return c_int32(result.value).value

"""
DEPRECATED: 旧的 SpecialNumber 算法
def calcSpecialNumber2():
    max = 1037933
    num2 = random.randint(1, max) * 2069

    num2 += 1024  # specialnum
    num3 = 0
    for i in range(0, 32):
        num3 <<= 1
        num3 += num2 % 2
        num2 >>= 1

    return num3
"""
